/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include <fcntl.h>
#include <ctype.h>
#include <sys/time.h>
#include "ftp_main.h"

/* include the different command sets */

#include "ftp_cmds.h"
#include "ftp_std.h"

static struct {
	struct cmd_handler_t handler[100];
	int nhandlers;
	const char* initial_expect;
	/* if this has the value 1, we wait for the command in initial_expect */
	int is_initial_state;
} cmd_handlers;

#define CMD_HANDLER_CMD_SEEN_AMOUNT 50
#define CMD_HANDLER_CMD_SEEN_LENGTH 20

static struct {
	char cmd_name[CMD_HANDLER_CMD_SEEN_LENGTH];
	int nseen;
} cmd_handler_cmd_seen[CMD_HANDLER_CMD_SEEN_AMOUNT];

struct cmdhandlerstruct *cmdhandler;
struct cmd_handler_t* cmd_get_next_handler(const char* cmd, int nargs,
						struct cmd_handler_t* lh);
static void cmd_decrease_seen(const char* cmd_name);

int handle_cmds(struct proxyinfo *pi, struct messagebuffer* msgbuf) {
	char* buffer = (char*) 0;
	char* msg;
	char** args = (char**) 0;
	int nargs;
	int i;
	int ss = pi->si->connection.serversocket;
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	struct serverinfo* si = pi->si;
	const char* DELIMITERS = "\t\n\r ";
	struct cmd_handler_t* ch;
	cmd_hookfunc_t err_reset = (cmd_hookfunc_t) 0;
	const char* expected = (const char*) 0;
	const char* expected_backup = (const char*) 0;
	int ret, final_action;
	int is_implementation;
	int is_handled;
	int expect_error;
	int respcode;
	int protoviolations = 0;

	/*
            Server host name                                     svrname
            Server IP address                                    svrip
            Server host name as specified in the login           svrlogin
            Client host name                                     clntname
            Client host IP address                               clntip
            Interface IP to the client                           ifipclnt
            Interface IP to the server                           ifipsvr
	*/
	cmd_handlers.is_initial_state = 1;

	while (1) {
contin:
		ss = si->connection.serversocket;

		if (check_say_timeout(cs, __FILE__, __LINE__)) {
			return -1;
		}

		errno = 0;
		buffer = readline(cs);

		if (!buffer) {
			if (check_say_timeout(cs, __FILE__, __LINE__)) {
				return -1;
			}
			if (!errno) {
				set_errstr("The client has probably closed the connection without saying QUIT");
			} else {
				set_errstr(strerror(errno));
			}
			err_readline(cs);
			return -1;
		}

		if (buffer[0] == '\0') {
			/* empty line. Prevent logging of the command */
			free(buffer);
			buffer = (char*) 0;
			goto contin;
		}

		args = quotstrtok_arr(buffer, DELIMITERS);
		if (strcasecmp(args[0], "PASS") == 0) {
			ci->lastcmd.cmd = "PASS *******";
		} else {
			ci->lastcmd.cmd = buffer;
		}
		ci->lastcmd.method = args[0];
		free_and_zero_char(&ci->lastcmd.filename);
		/* count the arguments */
		for (nargs = 0; args[nargs]; nargs++) {}
		/* the command is not an argument */
		nargs--;

		/* do an upper case of the command */
		toupstr(args[0]);

		/* We must not perform ASCII<->binary conversions if we are
		 * transferring a server list like LIST or NLST. Assume we
		 * don't have one and set it to 1 in the handler for the
		 * respective commands.
		 * XXX initial: 0, set in commands, reset after use
		 * */
		ci->data_connection.asciiconversion.serverlisting = 0;

		/* Default setting: neither up nor download */
		ci->data_connection.direction = TRANSFER_DIRECTION_UNSPEC;
		i = 0;

		expect_error = 0;
		if (cmd_handlers.is_initial_state
					&& cmd_handlers.initial_expect) {

			expected = cmd_handlers.initial_expect;
		}

		if (expected) {
			jlog(8, "Expecting %s", expected);
		} else {
			jlog(8, "No special command expected");
		}
		if (expected && strcmp(expected, args[0]) != 0
				/* this is not really clean but so far QUIT is
				 * the only command that gets allowed though it
				 * is not expected so I think this is easier
				 * than any other general solution that makes
				 * the commandhandler expect multiple commands
				 * (one being always QUIT) or marking commands
				 * saying that they are allowed to break the
				 * normal sequence... I'll do a better solution
				 * if I need one but this is fine for now... */
				&& strcmp(args[0], "QUIT") != 0) {
			/* this is NOT the expected command */
			jlog(6, "Got %s, expected %s", args[0], expected);
			message_buffer_addf(msgbuf, cs,
				500, "%s not understood. Expecting %s.\r\n",
							args[0], expected);
			jlog(5, "Command %s not expected (%s)", args[0], buffer);
			expect_error = 1;
		} else if (expected) {
			/* it matched, we got the expected command */
			/* we save a backup for the case that the command
			 * failed. Then we have to reset the expected variable
			 * such that it has the same value as before. */
			expected_backup = expected;
			expected = (const char*) 0;
		}
		if (my_strcasestr(buffer, "PASS ") == (char*) 0) {
			jlog(8, "Got: %s", buffer);
		} else {
			jlog(8, "Got the password");
		}

		ret = CMD_NOTFOUND;
		final_action = CMD_HANDLED;
		is_handled = 0;
		ch = (struct cmd_handler_t*) 0;
		while (!expect_error
			&& (ch = cmd_get_next_handler(args[0], nargs, ch))) {

			is_implementation = ch->priority >= CMD_HANDLERTYPE_RANGE_IMPLEMENT_MIN && ch->priority <= CMD_HANDLERTYPE_RANGE_IMPLEMENT_MAX;
			if (is_handled && is_implementation) {
				/* if we had a successful implementation
				 * handler already, skip every other
				 * implementation handler */
				continue;
			}
			expected_backup = (const char*) 0;
			if (ch->expect) {
				if (expected) {
					if (strcasecmp(expected, ch->expect)) {
						jlog(6, "expecting %s already, now %s\n",
								expected, ch->expect);
					}
				}
				/* XXX points to static memory? */
				expected = ch->expect;
			}
			err_reset = ch->err_reset;
			ASSUME(err_reset != NULL);
			ret = (ch->func)(args[0], (const char**) &args[1],
					pi, msgbuf);
			if (strcasecmp(args[0], "PASS ") == 0) {
				overwrite_string(buffer);
			}
			switch (ret) {
			case CMD_PREEXEC_OK:
				jlog(9, "cmd (%s) passed function %s",
						args[0], ch->description);
				break;
			case CMD_PREEXEC_FAIL:
				jlog(9, "cmd (%s) did not pass function %s",
						args[0], ch->description);
				/* the function has to send the
				 * reason */
				break;
			case CMD_POSTEXEC_OK:
				/* a postexec function can't really
				 * fail */
				break;
			case CMD_DONE:
				/* wait for Postexecs */
				final_action = CMD_DONE;
				is_handled = 1;
				break;
			case CMD_HANDLED:
				is_handled = 1;
				cmd_handlers.is_initial_state = 0;
				break;
			case CMD_NOT_HANDLED:
				break;
			case CMD_QUIT:
				transfer_cleanup(pi);
				/* wait for Postexecs */
				final_action = CMD_QUIT;
				is_handled = 1;
				ret = (err_reset) (args[0],
					(const char**) &args[1],
					pi, msgbuf);
				break;
			case CMD_ABORT:
				is_handled = 1;
				transfer_cleanup(pi);
				/* wait for Postexecs */
				final_action = CMD_ABORT;
				cmd_decrease_seen(ch->cmd_name);
				expected = (const char*) 0;
				ret = (err_reset) (args[0],
					(const char**) &args[1],
					pi, msgbuf);
				break;
			case CMD_ERROR:
				is_handled = 1;
				transfer_cleanup(pi);
				cmd_decrease_seen(ch->cmd_name);
				expected = (const char*) 0;
				cmd_handlers.is_initial_state = 1;
				/* reset especially counters for command
				 * sequels */
				cmd_handlers_reset();
				ret = (err_reset) (args[0],
					(const char**) &args[1],
					pi, msgbuf);
				break;
			}
			if (ret == CMD_PREEXEC_FAIL) {
				break;
			}
		}
		if (ret == CMD_NOTFOUND) {
			if (!expect_error) {
				message_buffer_addf(msgbuf, cs,
					500, "%s not understood\r\n", args[0]);
				jlog(7, "Command %s not understood (%s)",
							args[0], buffer);
				if (expected_backup) {
					expected = expected_backup;
					expected_backup = (const char*) 0;
				}
			} else {
				/* it got already logged */
			}
			protoviolations++;
			if (protoviolations >
			    config_get_ioption("protocolviolations", 10)){
				message_buffer_add(msgbuf, cs,
					500, "Too many consequent protocol "
					"violations - Closing connection\r\n");
				transfer_cleanup(pi);
				final_action = CMD_ABORT;
			}
		} else {
			if (is_handled == 0) {
				/* Set the error code to 0. If one of the
				 * handlers has set it, it will show up (the
				 * last). If not, the routine will use 500 as
				 * the default anyway... */
				message_buffer_addf(msgbuf, cs,
					0, "%s not understood\r\n", args[0]);
				jlog(5, "No handler succeeded for command "
						"%s (%s)", args[0], buffer);
			}
			protoviolations = 0;
		}
		/* Collect all the output for the client and send it in a
		 * single message */
		msg = message_buffer_getmsg_m(msgbuf, cs, &respcode);
		if (msg) {
			say(cs, msg);
			jlog(8, "Sending message to client: %s", msg);
			free(msg);
			ci->lastcmd.respcode = respcode;
		}
		message_buffer_reset(msgbuf);

		if (buffer) {
			if (is_handled) {
				/* log the command if it triggered an action */
				log_cmd(pi);
			}
			ci->lastcmd.method = (char*) 0;
			ci->lastcmd.respcode = 0;
			ci->data_connection.transferred = 0;
			free_and_zero_char(&buffer);
		}
		for (i = 0; args[i]; i++) {
			free_and_zero_char(&args[i]);
		}
		free(args);

		switch(final_action) {
			case CMD_DONE: /* XXX */
				jlog(8, "DONE!");
				expected = (char*) 0;
				break;
			case CMD_QUIT:
				jlog(8, "QUIT!");
				return 0;
			case CMD_ABORT:
				jlog(8, "ABORT!");
				return -1;
			default:
				break;
		}
	}

	return 0;
}

int transfer_cleanup(struct proxyinfo *pi) {
	struct clientinfo* ci = pi->ci;

	if (ci->data_connection.portcmd) {
		free_and_zero_char(&ci->data_connection.portcmd);
	}
	if (ci->data_connection.dataserversock != -1) {
		close(ci->data_connection.dataserversock);
	}
	if (ci->data_connection.dataclientsock != -1) {
		close(ci->data_connection.dataclientsock);
	}
	return 0;
}

static
void cmd_decrease_seen(const char* cmd_name) {
	int i;
	for (i = 0; i < CMD_HANDLER_CMD_SEEN_AMOUNT; i++) {
		if (strncmp(cmd_name, cmd_handler_cmd_seen[i].cmd_name,
					CMD_HANDLER_CMD_SEEN_LENGTH) == 0) {
			cmd_handler_cmd_seen[i].nseen--;
			return;
		}
	}
}

static
void cmd_increase_seen(const char* cmd_name) {
	int i;
	for (i = 0; i < CMD_HANDLER_CMD_SEEN_AMOUNT; i++) {
		if (strncmp(cmd_name, cmd_handler_cmd_seen[i].cmd_name,
				CMD_HANDLER_CMD_SEEN_LENGTH) == 0) {
			cmd_handler_cmd_seen[i].nseen++;
			return;
		}
	}
	/* if we are still here, it's a new command that is not yet in the
	 * table */
	for (i = 0; i < CMD_HANDLER_CMD_SEEN_AMOUNT; i++) {
		if (cmd_handler_cmd_seen[i].cmd_name[0] == '\0') {
			/* we have found a free entry */
			cmd_handler_cmd_seen[i].nseen++;
			strncpy(cmd_handler_cmd_seen[i].cmd_name, cmd_name,
					CMD_HANDLER_CMD_SEEN_LENGTH);
			return;
		}
	}
	if (i >= CMD_HANDLER_CMD_SEEN_AMOUNT) {
		/* the list of commands is full */
		jlog(1, "The list of commands is full. Can't count how often "
				"we've seen cmd '%s'", cmd_name);
	}
}

static
int cmd_get_seen(const char* cmd_name) {
	int i;
	for (i = 0; i < CMD_HANDLER_CMD_SEEN_AMOUNT; i++) {
		if (strncmp(cmd_name, cmd_handler_cmd_seen[i].cmd_name,
					CMD_HANDLER_CMD_SEEN_LENGTH) == 0) {
			return cmd_handler_cmd_seen[i].nseen;
		}
	}
	return -1;
}

int cmd_dummy_reset(const char* cmd, const char** args,
		struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	jlog(8, "dummy reset function - no reset function defined");
	return 0;
}

int cmd_register(int priority,
		 int serialize,
		 const char* cmd_name,
		 const char* description,
		 int ncmd_args_min,
		 int ncmd_args_max,
		 int serial_id,
		 cmd_hookfunc_t hookfunc,
		 const char* expect,
		 cmd_hookfunc_t err_reset) {

	int i;

	if (cmd_handlers.nhandlers > 99) {
		/* XXX: Error */
		jlog(2, "Too many commands registered.");
		return -1;
	}

	/* check for serialize options */
	if (serialize == CMD_SERIALIZE_REMOVE_ONE) {
		int j;
		for (j = 0; j < cmd_handlers.nhandlers; j++) {
			if (strcasecmp(cmd_handlers.handler[j].cmd_name,
						cmd_name) == 0
				&& cmd_handlers.handler[j].priority
						== priority
				&& cmd_handlers.handler[j].serial_id
						== serial_id) {
				/* remove this one */
				cmd_handlers.handler[j] = cmd_handlers.handler[cmd_handlers.nhandlers - 1];
				cmd_handlers.nhandlers--;
			}
		}
	}
	if (serialize == CMD_SERIALIZE_REMOVE_ALL) {
		int j;
		for (j = 0; j < cmd_handlers.nhandlers; j++) {
			if (strcasecmp(cmd_handlers.handler[j].cmd_name,
						cmd_name) == 0
				&& cmd_handlers.handler[j].serial_id
						== serial_id) {
				/* remove this one - no matter what priority
				 * it has */
				cmd_handlers.nhandlers--;
				cmd_handlers.handler[j] = cmd_handlers.handler[cmd_handlers.nhandlers];
			}
		}
	}

	i = cmd_handlers.nhandlers;
	/* if we have nhandlers handlers, cmd_handlers[nhandlers] is the
	 * next array element to be filled */
	cmd_handlers.handler[i].priority = priority;
	cmd_handlers.handler[i].cmd_name = cmd_name;
	cmd_handlers.handler[i].description = description;
	cmd_handlers.handler[i].ncmd_args_min = ncmd_args_min;
	cmd_handlers.handler[i].ncmd_args_max = ncmd_args_max;
	cmd_handlers.handler[i].func = hookfunc;
	cmd_handlers.handler[i].serial_id = serial_id;
	cmd_handlers.handler[i].expect = expect;
	if (err_reset) {
		cmd_handlers.handler[i].err_reset = err_reset;
	} else {
		cmd_handlers.handler[i].err_reset = cmd_dummy_reset;
	}

	cmd_handlers.nhandlers++;

	return 0;
}

void cmd_set_initial_expect(const char* expected) {
	cmd_handlers.initial_expect = expected;
}

void cmd_handlers_reset() {
	int i;

	for (i = 0; i < CMD_HANDLER_CMD_SEEN_AMOUNT; i++) {
		cmd_handler_cmd_seen[i].nseen = -1;
		cmd_handler_cmd_seen[i].cmd_name[0] = '\0';
	}
}

void cmd_handlers_init() {
	cmd_handlers.initial_expect = (char*) 0;
	cmd_handlers.is_initial_state = 0;
	cmd_handlers.nhandlers = 0;
}

static
int cmd_init_config_forward_cmds(void) {
	const char* cmds = config_get_option("forwardcmds");
	char* cmd;
	int offset = 0;
	if (!cmds) {
		return 0;
	}

	while((cmd = quotstrtok(cmds, WHITESPACE, &offset))) {
		/* XXX-MEM cmd won't be freed */
		cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
				CMD_SERIALIZE_REMOVE_ONE,
				/* say 5 here as maximal parameter count.
				 * Problems with that? */
				cmd, "forwarded by config file", 0, 5,
				-1,
				std_config_forward_cmd,
				(char*) 0, (cmd_hookfunc_t) 0);
		/* we can NOT free cmd, since the address is still listed in
		 * the command handler */
	}
	return 0;
}

int std_cmds_init() {
	cmd_handlers_init();
	cmd_handlers_reset();

	/* register all the commands */
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"QUIT", "standard QUIT", 0, 0,
			-1,
			std_quit,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASV", "standard PASV", 0, 0,
			-1,
			std_pasv,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PORT", "standard PORT", 1, 1,
			-1,
			std_port,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASS", "standard PASS", -1, -1,
			-1,
			std_already_loggedin,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"STOR", "standard STOR", 1, 1,
			-1,
			std_stor,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"STOU", "standard STOU", 0, 0,
			-1,
			std_stou,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"APPE", "standard APPE", 1, 1,
			-1,
			std_stor,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"TYPE", "standard TYPE", 1, 1,
			-1,
			std_type,
			(char*) 0, (cmd_hookfunc_t) 0);

	/* I don't know if 1 is really the maximal number of arguments.
	 * However I don't know a program that uses more than one... */
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"LIST", "standard LIST", 0, 1,
			-1,
			std_list,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"NLST", "standard NLST", 0, 1,
			-1,
			std_list,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"RETR", "standard RETR", 1, 1,
			-1,
			std_retr,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"CAPA", "List Capabilities", 0, 0,
			-1,
			std_capa,
			(char*) 0, (cmd_hookfunc_t) 0);

	/* ------- commands that we just forward ------- */
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"CWD", "standard CWD", 1, 1,
			-1,
			std_cwd,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PWD", "standard PWD", 0, 0,
			-1,
			std_pwd,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"CDUP", "standard CDUP", 0, 0,
			-1,
			std_cdup,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"XCUP", "standard XCUP", 0, 0,
			-1,
			std_cdup,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"SIZE", "standard SIZE", 1, 1,
			-1,
			std_size,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"MKD", "standard MKD", 1, 1,
			-1,
			std_mkd,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"XMKD", "standard XMKD", 1, 1,
			-1,
			std_mkd,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"RMD", "standard RMD", 1, 1,
			-1,
			std_rmd,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"XRMD", "standard XRMD", 1, 1,
			-1,
			std_rmd,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"RNFR", "standard RNFR", 1, 1,
			-1,
			std_rnfr,
			/* XXX */
			"RNTO", (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"RNTO", "standard RNTO", 1, 1,
			-1,
			std_rnto,
			(char*) 0, (cmd_hookfunc_t) 0);
	
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"NOOP", "standard NOOP", 0, 0,
			-1,
			std_noop,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"MDTM", "standard MDTM", 1, 1,
			-1,
			std_mdtm,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"REST", "standard REST", 1, 1,
			-1,
			std_rest,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"DELE", "standard DELE", 1, 1,
			-1,
			std_dele,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			/* I've put 0..5 here because it's really server
			 * dependent... */
			"SITE", "standard SITE", 0, 5,
			-1,
			std_site,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_init_config_forward_cmds();
	
	return 0;
}

static
int handler_priority_compare(const void* a_v, const void* b_v) {
	const struct cmd_handler_t *a = *(const struct cmd_handler_t**) a_v;
	const struct cmd_handler_t *b = *(const struct cmd_handler_t**) b_v;

	if (a->priority < b->priority) {
		return -1;
	}
	if (a->priority > b->priority) {
		return 1;
	}
	return 0;
}

struct cmd_handler_t* cmd_get_next_handler(const char* cmd, int nargs,
						struct cmd_handler_t* lh) {
	/* XXX-MEM the last one won't be freed */
	static struct cmd_handler_t** mh;
	static int returned;
	static int mh_reserved;
	int i, j;
	int serial_id = 0;
	int implemented = 0;  /* assume we don't have an implementation */

	if (!mh_reserved) {
		mh_reserved = 3;
		mh = malloc(sizeof(struct cmd_handler_t*) * mh_reserved);
		enough_mem(mh);
	}

	if (lh == (struct cmd_handler_t*) 0) {
		/* create new list of command handlers, including preexec,
		 * implemented and postexec handlers */
		returned = 0;
		j = 0;
		serial_id = cmd_get_seen(cmd) + 1;
		/* i runs on the whole list of command handlers, j on our
		 * part where we grab the pointers to the handlers which are
		 * of interest */
		for (i = 0; i < cmd_handlers.nhandlers; i++) {
			const char* c_name = cmd_handlers.handler[i].cmd_name;
			int priority = cmd_handlers.handler[i].priority;
			if ((strcasecmp(c_name, cmd) == 0
					|| strcmp(c_name, "*") == 0)
				&& (cmd_handlers.handler[i].ncmd_args_min <= nargs
					|| cmd_handlers.handler[i].ncmd_args_min == -1)
				&& (cmd_handlers.handler[i].ncmd_args_max >= nargs
					|| cmd_handlers.handler[i].ncmd_args_max == -1)
				&& (cmd_handlers.handler[i].serial_id == serial_id
					|| cmd_handlers.handler[i].serial_id == -1)) {

				/* of interest */
				if (j == mh_reserved - 1) {
					/* leave one for a terminating NULL
					 * entry */
					mh_reserved++;
					mh = realloc(mh, sizeof(struct cmd_handler_t*) * mh_reserved);
				}
				mh[j] = &(cmd_handlers.handler[i]);
				j++;
				qsort(&mh[0], j, sizeof(struct cmd_handler_t*),
						handler_priority_compare);
				if (priority >= CMD_HANDLERTYPE_RANGE_IMPLEMENT_MIN && priority <= CMD_HANDLERTYPE_RANGE_IMPLEMENT_MAX) {
					implemented = 1;
				}
			}
		}
		mh[j] = (struct cmd_handler_t*) 0;
		if (implemented) {
			if (strcmp(mh[0]->cmd_name, "*") != 0) {
				/* there is no sense in defining a serie of
				 * different catchall handlers */
				cmd_increase_seen(mh[0]->cmd_name);
			}
			return mh[0];
		} else {
			/* maybe we have Pre- or Postexecs but no
			 * implementation. Don't accept the command */
			return (struct cmd_handler_t*) 0;
		}
	} else {
		/* we have not yet returned all the handlers, return the next
		 * one from the list previously created. */
		if (lh != mh[returned]) {
			/* XXX: error! */
		} else {
			returned++;
			if (returned == mh_reserved) {
				return (struct cmd_handler_t*) 0;
			}
			if (mh[returned]) {
				cmd_increase_seen(mh[returned]->cmd_name);
			}
			/* we return a null pointer the last time */
			return mh[returned];
		}
	}
	return (struct cmd_handler_t*) 0;
}


